package edu.rice.comp322;

import java.util.ArrayList;
import java.util.List;

/**
 * This class should not be modified.
 *
 * @author <a href="http://shams.web.rice.edu/">Shams Imam</a> (shams@rice.edu)
 */
public final class ProblemState {

    protected final int numMoves;
    protected final int numValues;
    protected final int boardSize;
    protected final List<Integer>[] feasibleValueTable;
    protected final List<Integer> mutatedLocations;
    protected final int numUnsolved;

    protected ProblemState(final int[] setIntegers) {

        SudokuConfig.incrementCount();
        this.numMoves = 0;

        // compute noRows = noColumns = boardSize, knowing that: numValues = boardSize * boardSize (square board)
        this.numValues = setIntegers.length;
        this.boardSize = (int) Math.sqrt(numValues);
        this.mutatedLocations = new ArrayList<Integer>();

        @SuppressWarnings("unchecked")
        final List<Integer>[] temp = (List<Integer>[]) new List[numValues];
        for (int i = 0; i < numValues; i++) {
            temp[i] = new ArrayList<>();
        }
        for (int i = 0; i < numValues; i++) {
            final int setValue = setIntegers[i];
            if (setValue != -1) {
                temp[i].add(setValue);
                mutatedLocations.add(i);
            }
        }
        for (int i = 0; i < numValues; i++) {
            if (temp[i].isEmpty()) {
                final List<Integer> possibleValues = SudokuUtils.fillPossibleValues(i, boardSize, temp);
                if (possibleValues.size() == 0) {
                    throw new IllegalStateException("Invalid board!");
                }
                temp[i].addAll(possibleValues);
            }
        }
        this.feasibleValueTable = temp;
        this.numUnsolved = SudokuUtils.countUnsolved(this.feasibleValueTable);
    }

    private ProblemState(final int numMoves, final List<Integer>[] setIntegers) {

        SudokuConfig.incrementCount();
        this.numMoves = numMoves;

        // compute noRows = noColumns = boardSize, knowing that: numValues = boardSize * boardSize (square board)
        this.numValues = setIntegers.length;
        this.boardSize = (int) Math.sqrt(numValues);
        this.mutatedLocations = new ArrayList<Integer>();

        @SuppressWarnings("unchecked")
        final List<Integer>[] temp = (List<Integer>[]) new List[numValues];
        for (int i = 0; i < numValues; i++) {
            temp[i] = new ArrayList<>();
        }
        for (int i = 0; i < numValues; i++) {
            temp[i].addAll(setIntegers[i]);
        }
        this.feasibleValueTable = temp;
        this.numUnsolved = SudokuUtils.countUnsolved(this.feasibleValueTable);
    }

    protected List<ProblemState> neighbors() {

        final List<ProblemState> resultList = new ArrayList<>();

        int indexToFocusOn = -1;
        for (int i = 0; i < feasibleValueTable.length; i++) {
            final List<Integer> valueHolders = feasibleValueTable[i];
            if (valueHolders.size() > 1) {
                indexToFocusOn = i;
                break;
            }
        }

        if (indexToFocusOn >= 0) {

            final List<Integer> possibleValues = feasibleValueTable[indexToFocusOn];
            for (final Integer possibleValue : possibleValues) {
                final ProblemState newBoard = setPossibleValue(indexToFocusOn, possibleValue);

                // System.out.println(newBoard);

                if (newBoard.isConsistent()) {
                    resultList.add(newBoard);
                }
            }

        }

        return resultList;
    }

    protected ProblemState setPossibleValue(final int location, final int value) {
        final ProblemState newBoard = new ProblemState(this.numMoves + 1, this.feasibleValueTable);
        final List<Integer>[] newFvt = newBoard.feasibleValueTable;
        final List<Integer> newMutatedLocs = newBoard.mutatedLocations;
        SudokuUtils.setPossibleValue(location, value, boardSize, newFvt, newMutatedLocs);
        return newBoard;
    }

    protected boolean isSolution() {
        return SudokuUtils.isSolution(numValues, feasibleValueTable);
    }

    protected boolean isConsistent() {
        return SudokuUtils.isConsistent(numValues, feasibleValueTable, mutatedLocations);
    }

    @Override
    public boolean equals(final Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        final ProblemState other = (ProblemState) o;
        return this.boardString().equals(other.boardString());
    }

    @Override
    public int hashCode() {
        int result = numMoves;
        result = 31 * result + numValues;
        result = 31 * result + boardSize;
        result = 31 * result + numUnsolved;
        return result;
    }

    @Override
    public String toString() {
        final StringBuilder s = new StringBuilder();

        s.append("* State [moves=");
        s.append(numMoves);
        s.append(", unsolved=");
        s.append(numUnsolved);
        s.append("]: \n");

        s.append(boardString());

        return s.toString();
    }

    protected String boardString() {

        final StringBuilder s = new StringBuilder(numValues * 2 + boardSize);

        final int rows = boardSize;
        for (int r = 0; r < rows; r++) {
            for (int c = 0; c < rows; c++) {

                final List<Integer> integers = feasibleValueTable[r * rows + c];

                if (integers.size() == 1) {
                    final int intValue = integers.get(0);
                    if (intValue < 10) {
                        s.append(" ").append(intValue);
                    } else {
                        final char charValue = (char) (intValue - 10 + 'A');
                        s.append(" ").append(charValue);
                    }
                } else if (integers.size() > 1) {
                    s.append(" .");
                } else {
                    s.append(" X");
                }
            }
            s.append("\n");
        }

        return s.toString();
    }
}
